@@ -22,7 +22,9 @@ import java.nio.charset.StandardCharsets;
 import java.security.AccessController;
 import java.security.PrivilegedActionException;
 import java.security.PrivilegedExceptionAction;
+import java.util.HashSet;
 import java.util.Locale;
+import java.util.Set;
 
 import org.apache.coyote.ActionCode;
 import org.apache.coyote.CloseNowException;
@@ -31,6 +33,7 @@ import org.apache.coyote.Request;
 import org.apache.coyote.Response;
 import org.apache.coyote.http11.HttpOutputBuffer;
 import org.apache.coyote.http11.OutputFilter;
+import org.apache.coyote.http11.filters.SavedRequestInputFilter;
 import org.apache.coyote.http2.HpackDecoder.HeaderEmitter;
 import org.apache.juli.logging.Log;
 import org.apache.juli.logging.LogFactory;
@@ -56,11 +59,19 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
 
     private static final Integer HTTP_UPGRADE_STREAM = Integer.valueOf(1);
 
+    private static final Set<String> HTTP_CONNECTION_SPECIFIC_HEADERS = new HashSet<>();
+
     static {
         Response response =  new Response();
         response.setStatus(100);
         StreamProcessor.prepareHeaders(null, response, null, null);
         ACK_HEADERS = response.getMimeHeaders();
+
+        HTTP_CONNECTION_SPECIFIC_HEADERS.add("connection");
+        HTTP_CONNECTION_SPECIFIC_HEADERS.add("proxy-connection");
+        HTTP_CONNECTION_SPECIFIC_HEADERS.add("keep-alive");
+        HTTP_CONNECTION_SPECIFIC_HEADERS.add("transfer-encoding");
+        HTTP_CONNECTION_SPECIFIC_HEADERS.add("upgrade");
     }
 
     private volatile long contentLengthReceived = 0;
@@ -78,6 +89,7 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
     private StreamException headerException = null;
 
     private volatile StringBuilder cookieHeader = null;
+    private volatile boolean hostHeaderSeen = false;
 
     private Object pendingWindowUpdateForStreamLock = new Object();
     private int pendingWindowUpdateForStream = 0;
@@ -96,12 +108,13 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
         if (coyoteRequest == null) {
             // HTTP/2 new request
             this.coyoteRequest = new Request();
-            this.inputBuffer = new StreamInputBuffer();
+            this.inputBuffer = new StandardStreamInputBuffer();
             this.coyoteRequest.setInputBuffer(inputBuffer);
         } else {
             // HTTP/2 Push or HTTP/1.1 upgrade
             this.coyoteRequest = coyoteRequest;
-            this.inputBuffer = null;
+            this.inputBuffer = new SavedRequestStreamInputBuffer(
+                    (SavedRequestInputFilter) coyoteRequest.getInputBuffer());
             // Headers have been read by this point
             state.receivedStartOfHeaders();
             if (HTTP_UPGRADE_STREAM.equals(identifier)) {
@@ -137,7 +150,7 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
             throw new IllegalArgumentException();
         }
         // This processing expects bytes. Server push will have used a String
-        // to trigger a conversion if required.
+        // so trigger a conversion if required.
         hostValueMB.toBytes();
         ByteChunk valueBC = hostValueMB.getByteChunk();
         byte[] valueB = valueBC.getBytes();
@@ -281,9 +294,9 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
                     getConnectionId(), getIdAsString(), name));
         }
 
-        if ("connection".equals(name)) {
+        if (HTTP_CONNECTION_SPECIFIC_HEADERS.contains(name)) {
             throw new HpackException(sm.getString("stream.header.connection",
-                    getConnectionId(), getIdAsString()));
+                    getConnectionId(), getIdAsString(), name));
         }
 
         if ("te".equals(name)) {
@@ -365,20 +378,7 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
         }
         case ":authority": {
             if (coyoteRequest.serverName().isNull()) {
-                int i;
-                try {
-                    i = Host.parse(value);
-                } catch (IllegalArgumentException iae) {
-                    // Host value invalid
-                    throw new HpackException(sm.getString("stream.header.invalid",
-                            getConnectionId(), getIdAsString(), ":authority", value));
-                }
-                if (i > -1) {
-                    coyoteRequest.serverName().setString(value.substring(0, i));
-                    coyoteRequest.setServerPort(Integer.parseInt(value.substring(i + 1)));
-                } else {
-                    coyoteRequest.serverName().setString(value);
-                }
+                parseAuthority(value, false);
             } else {
                 throw new HpackException(sm.getString("stream.header.duplicate",
                         getConnectionId(), getIdAsString(), ":authority" ));
@@ -396,6 +396,22 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
             cookieHeader.append(value);
             break;
         }
+        case "host": {
+            if (coyoteRequest.serverName().isNull()) {
+                // No :authority header. This is first host header. Use it.
+                hostHeaderSeen = true;
+                parseAuthority(value, true);
+            } else if (!hostHeaderSeen) {
+                // First host header - must be consistent with :authority
+                hostHeaderSeen = true;
+                compareAuthority(value);
+            } else {
+                // Multiple hosts headers - illegal
+                throw new HpackException(sm.getString("stream.header.duplicate",
+                        getConnectionId(), getIdAsString(), "host" ));
+            }
+            break;
+        }
         default: {
             if (headerState == HEADER_STATE_TRAILER && !handler.isTrailerHeaderAllowed(name)) {
                 break;
@@ -415,6 +431,45 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
     }
 
 
+    private void parseAuthority(String value, boolean host) throws HpackException {
+        int i;
+        try {
+            i = Host.parse(value);
+        } catch (IllegalArgumentException iae) {
+            // Host value invalid
+            throw new HpackException(sm.getString("stream.header.invalid",
+                    getConnectionId(), getIdAsString(), host ? "host" : ":authority", value));
+        }
+        if (i > -1) {
+            coyoteRequest.serverName().setString(value.substring(0, i));
+            coyoteRequest.setServerPort(Integer.parseInt(value.substring(i + 1)));
+        } else {
+            coyoteRequest.serverName().setString(value);
+        }
+    }
+
+
+    private void compareAuthority(String value) throws HpackException {
+        int i;
+        try {
+            i = Host.parse(value);
+        } catch (IllegalArgumentException iae) {
+            // Host value invalid
+            throw new HpackException(sm.getString("stream.header.invalid",
+                    getConnectionId(), getIdAsString(), "host", value));
+        }
+        if (i == -1 && (!value.equals(coyoteRequest.serverName().getString()) || coyoteRequest.getServerPort() != -1) ||
+                i > -1 && ((!value.substring(0, i).equals(coyoteRequest.serverName().getString()) ||
+                    Integer.parseInt(value.substring(i + 1)) != coyoteRequest.getServerPort()))) {
+            // Host value inconsistent
+            throw new HpackException(sm.getString("stream.host.inconsistent",
+                    getConnectionId(), getIdAsString(), value, coyoteRequest.serverName().getString(),
+                    Integer.toString(coyoteRequest.getServerPort())));
+        }
+
+    }
+
+
     @Override
     public void setHeaderException(StreamException streamException) {
         if (headerException == null) {
@@ -824,7 +879,7 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
         @Override
         public final synchronized int doWrite(ByteBuffer chunk) throws IOException {
             if (closed) {
-                throw new IllegalStateException(
+                throw new IOException (
                         sm.getString("stream.closed", getConnectionId(), getIdAsString()));
             }
             int totalThisTime = 0;
@@ -1008,7 +1063,27 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
     }
 
 
-    class StreamInputBuffer implements InputBuffer {
+    abstract class StreamInputBuffer implements InputBuffer {
+
+        abstract void receiveReset();
+
+        abstract void swallowUnread() throws IOException;
+
+        abstract void notifyEof();
+
+        abstract ByteBuffer getInBuffer();
+
+        abstract void onDataAvailable() throws IOException;
+
+        abstract boolean isReadyForRead();
+
+        abstract boolean isRequestBodyFullyRead();
+
+        abstract void insertReplayedBody(ByteChunk body);
+    }
+
+
+    class StandardStreamInputBuffer extends StreamInputBuffer {
 
         /* Two buffers are required to avoid various multi-threading issues.
          * These issues arise from the fact that the Stream (or the
@@ -1202,6 +1277,7 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
         }
 
 
+        @Override
         final boolean isReadyForRead() {
             ensureBuffersExist();
 
@@ -1218,6 +1294,7 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
             }
         }
 
+        @Override
         final synchronized boolean isRequestBodyFullyRead() {
             return (inBuffer == null || inBuffer.position() == 0) && isInputFinished();
         }
@@ -1235,6 +1312,7 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
         /*
          * Called after placing some data in the inBuffer.
          */
+        @Override
         final synchronized void onDataAvailable() throws IOException {
             if (closed) {
                 swallowUnread();
@@ -1259,12 +1337,14 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
         }
 
 
-        private final ByteBuffer getInBuffer() {
+        @Override
+        final ByteBuffer getInBuffer() {
             ensureBuffersExist();
             return inBuffer;
         }
 
 
+        @Override
         final synchronized void insertReplayedBody(ByteChunk body) {
             inBuffer = ByteBuffer.wrap(body.getBytes(),  body.getOffset(),  body.getLength());
         }
@@ -1286,7 +1366,8 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
         }
 
 
-        private final void receiveReset() {
+        @Override
+        final void receiveReset() {
             if (inBuffer != null) {
                 synchronized (inBuffer) {
                     resetReceived = true;
@@ -1295,7 +1376,8 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
             }
         }
 
-        private final void notifyEof() {
+        @Override
+        final void notifyEof() {
             if (inBuffer != null) {
                 synchronized (inBuffer) {
                     inBuffer.notifyAll();
@@ -1303,7 +1385,8 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
             }
         }
 
-        private final void swallowUnread() throws IOException {
+        @Override
+        final void swallowUnread() throws IOException {
             synchronized (this) {
                 closed = true;
             }
@@ -1329,4 +1412,71 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
             }
         }
     }
+
+
+    class SavedRequestStreamInputBuffer extends StreamInputBuffer {
+
+        private final SavedRequestInputFilter inputFilter;
+
+        SavedRequestStreamInputBuffer(SavedRequestInputFilter inputFilter) {
+            this.inputFilter = inputFilter;
+        }
+
+
+        @Override
+        public int doRead(ApplicationBufferHandler handler) throws IOException {
+            return inputFilter.doRead(handler);
+        }
+
+        @Deprecated
+        @Override
+        public int doRead(ByteChunk chunk) throws IOException {
+            return inputBuffer.doRead(chunk);
+        }
+
+        @Override
+        public int available() {
+            return inputFilter.available();
+        }
+
+        @Override
+        void receiveReset() {
+            // NO-OP
+        }
+
+        @Override
+        void swallowUnread() throws IOException {
+            // NO-OP
+        }
+
+        @Override
+        void notifyEof() {
+            // NO-OP
+        }
+
+        @Override
+        ByteBuffer getInBuffer() {
+            return null;
+        }
+
+        @Override
+        void onDataAvailable() throws IOException {
+            // NO-OP
+        }
+
+        @Override
+        boolean isReadyForRead() {
+            return true;
+        }
+
+        @Override
+        boolean isRequestBodyFullyRead() {
+            return inputFilter.isFinished();
+        }
+
+        @Override
+        void insertReplayedBody(ByteChunk body) {
+            // NO-OP
+        }
+    }
 }
